/*
						+--------------------------------+
                        |                                |
                        |     *** TX2 Compiller ***      |
                        |                                |
                        |          L.Polok 2002         |
                        |                                |
                        |             Main.c             |
                        |                                |
						+--------------------------------+
*/

#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "ImgLoader.h"

typedef struct {
	char type[3];
	short version;
	short pal_num;
	short mip_num;
	short frm_num;
	short compression;
	short TWidth, THeight;
	short tex_type;
} TEXTURE_HEADER;

typedef struct {
	unsigned char chunk_hdr[3];
	char *chunk;
} FUNCTION_CHUNK;

BMP *Load_Tx2(char *filename)
{
	unsigned __int32 *pal;
	TEXTURE_HEADER thdr;
	FUNCTION_CHUNK *ch;
    BMP *tmp;
	FILE *fr;
	int i, j;

	if((fr = fopen(filename, "rb")) == NULL)
		return NULL;
	fread(&thdr, sizeof(TEXTURE_HEADER), 1, fr);  
	// read header

	if(thdr.type[0] != 't' ||
	   thdr.type[1] != 'x' ||
	   thdr.type[2] != '2')
		return NULL;
	if(thdr.version != 100)
		return NULL;

	if((pal = (unsigned __int32*) malloc(thdr.pal_num * 4)) == NULL)
		return NULL;
	if((ch = (FUNCTION_CHUNK*) malloc(thdr.pal_num * sizeof(FUNCTION_CHUNK))) == NULL)
		return NULL;
	if((tmp = (BMP*) malloc(sizeof(BMP))) == NULL)
		return NULL;
	if((tmp->bytes = (unsigned __int32*) malloc
	   (4 * thdr.TWidth * thdr.THeight)) == NULL)
		return NULL;
	tmp->Width = thdr.TWidth;
	tmp->Height = thdr.THeight;

	for(i = 0; i < thdr.pal_num; i ++)
		fread(&pal[i], 4, 1, fr);

	__int16 int16[1];
	__int8 int8[1];
	unsigned char chunk_hdr[3];
	char *chunk;

	int chunk_len;
	int var_num;
	int unit_cnt;

	long *val, *neg;
	long nval, nneg;
	int r, k, bit;
	int bit_v;

	if(thdr.compression != 1) {
		if(thdr.pal_num > 256) {
			for(i = 0; i < tmp->Width * tmp->Height; i ++) {
				fread(int16, 2, 1, fr);
				tmp->bytes[i] = pal[int16[0]];
			}
		} else {
			for(i = 0; i < tmp->Width * tmp->Height; i ++) {
				fread(int8, 1, 1, fr);
				tmp->bytes[i] = pal[int8[0]];
			}
		}
	} else {
		for(i = 0; i < thdr.pal_num; i ++) {
			fread(&ch[i].chunk_hdr[0], 1, 1, fr);
			fread(&ch[i].chunk_hdr[1], 1, 1, fr);
			fread(&ch[i].chunk_hdr[2], 1, 1, fr);
			if(ch[i].chunk_hdr[0] == 0 &&
			   ch[i].chunk_hdr[1] == 0 &&
			   ch[i].chunk_hdr[2] == 0)
			    continue; // unused color
			if(ch[i].chunk_hdr[0] == 0x80 &&
			   ch[i].chunk_hdr[1] == 0 &&
			   ch[i].chunk_hdr[2] == 0) {
				continue;
			} // bkgnd

			chunk_len = (ch[i].chunk_hdr[0] << 8) + ch[i].chunk_hdr[1];
			var_num = ch[i].chunk_hdr[2];

			ch[i].chunk = (char*) malloc(chunk_len);
			for(j = 0; j < chunk_len; j ++) 
				fread(&ch[i].chunk[j], 1, 1, fr);
		}
		// read chunks
		for(i = thdr.pal_num - 1; i >= 0; i --) {
			if(ch[i].chunk_hdr[0] == 0 &&
			   ch[i].chunk_hdr[1] == 0 &&
			   ch[i].chunk_hdr[2] == 0)
			    continue; // unused color
			if(ch[i].chunk_hdr[0] == 0x80 &&
			   ch[i].chunk_hdr[1] == 0 &&
			   ch[i].chunk_hdr[2] == 0) {
				for(j = 0; j < tmp->Width * tmp->Height; j ++)
					tmp->bytes[j] = pal[i];
				continue;
			} // bkgnd
			var_num = ch[i].chunk_hdr[2];
			chunk_len = (ch[i].chunk_hdr[0] << 8) + ch[i].chunk_hdr[1];
			unit_cnt = (int)(chunk_len * 4 / var_num);
				
			val = (long*) malloc(unit_cnt * sizeof(long));
			neg = (long*) malloc(unit_cnt * sizeof(long));

			/*
chunk[3 + ((j + k * (var_num << 1)) >> 3)] |=
 ((tmp_unit[j >> 3] & (1 << (j & 7))) << ((j + k * (var_num << 1)) & 7));
  */

			for(j = 0; j < unit_cnt; j ++) {
				val[j] = 0;
				neg[j] = 0;
				for(k = 0; k < var_num; k ++) {
					bit = k + j * (var_num << 1);
					bit_v = ((ch[i].chunk[bit >> 3] >> (bit & 7)) & 1);
					neg[j] |= bit_v << (var_num - 1 - k);
					bit = k + j * (var_num << 1) + var_num;
					bit_v = ((ch[i].chunk[bit >> 3] >> (bit & 7)) & 1);
					val[j] |= bit_v << (var_num - 1 - k);
				}
			}
			for(j = 0; j < tmp->Width * tmp->Height; j ++) {
				r = 0;
				for(k = 0; k < unit_cnt; k ++) {
					nval = (unsigned)j & (unsigned)val[k];
					nneg = (~(unsigned)j) & (unsigned)neg[k];
					r |= (nval == val[k]) && (nneg == neg[k]);
					//r |= ((((unsigned)j & (unsigned)val[k]) == (unsigned)val[k]) ? (unsigned)val[k] : 0) +
					//	((((unsigned char)(~j) & (unsigned)neg[k]) == (unsigned)neg[k]) ? (unsigned)neg[k] : 0);
				}
				if(r)
					tmp->bytes[j] = pal[i];
			}

			free((void*)ch[i].chunk);

			free((void*)val);
			free((void*)neg);
		}
		// translate chunks
	}

	free((void*)ch);
	free((void*)pal);

	/*
txhdr.version = 100;
		txhdr.pal_num = pal->Val_Num();
		txhdr.mip_num = mip_cnt;
		txhdr.frm_num = bmp_cnt;
		txhdr.TWidth = bmp_list[0]->Width;
		txhdr.THeight = bmp_list[0]->Height;
		txhdr.compression = compress;
		txhdr.tex_type = 0; // transparency, masking, etc..
		fwrite(&txhdr, sizeof(txhdr), 1, fw);
  */

	fclose(fr);

	return tmp;
}

void Print_Logo()
{
	printf("                    ____Tx2Tester v1.0____ \n");
	printf("                      Supports .bmp .pcx   \n");
	printf("                      and true-color bmp    \n\n\n");
}

void main(int argc, char **argv)
{
	CImgLoader *ildr;

	Print_Logo();
	// logo

	ildr = new CImgLoader;
	// init

	BMP *bmp;

	bmp = Load_Tx2("pcx_uncompr.tx2");
	if(bmp != NULL)
		ildr->Save_TrueColor_BMP("tx2.bmp", bmp);
	bmp = Load_Tx2("pcx_compr.tx2");
	if(bmp != NULL)
		ildr->Save_TrueColor_BMP("tx2_compr.bmp", bmp);
	// create texture

	delete ildr;
	// uninit

	getchar();

}


 

